""" overridable filter object to inject fields to auto-filter upon within searches """
from __future__ import absolute_import
from datetime import datetime

from django.conf import settings

from .utils import _load_class, DateRange


class SearchFilterGenerator(object):

    """
    Class to provide a set of filters for the search.
    Users of this search app will override this class and update setting for SEARCH_FILTER_GENERATOR
    """

    # disabling pylint violations because overriders will want to use these
    # pylint: disable=unused-argument, no-self-use
    def filter_dictionary(self, **kwargs):
        """ base implementation which filters via start_date """
        return {"start_date": DateRange(None, datetime.utcnow())}

    def field_dictionary(self, **kwargs):
        """ base implementation which add course if provided """
        field_dictionary = {}
        if "course_id" in kwargs and kwargs["course_id"]:
            field_dictionary["course"] = kwargs["course_id"]

        return field_dictionary

    def exclude_dictionary(self, **kwargs):
        """ base implementation which excludes nothing """
        return {}

    @classmethod
    def generate_field_filters(cls, **kwargs):
        """
        Called from within search handler
        Finds desired subclass and adds filter information based upon user information
        """
        generator = _load_class(getattr(settings, "SEARCH_FILTER_GENERATOR", None), cls)()
        return (
            generator.field_dictionary(**kwargs),
            generator.filter_dictionary(**kwargs),
            generator.exclude_dictionary(**kwargs),
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Some of the subclasses that get used as settings-overrides will yield this pylint
# error, but they do get used when included as part of the override_settings
# pylint: disable=too-few-public-methods
# pylint: disable=too-many-ancestors
""" Tests for search functionalty """
from __future__ import absolute_import
from datetime import datetime

from django.core.cache import cache
from django.test import TestCase
from django.test.utils import override_settings
from elasticsearch import Elasticsearch

from search.search_engine_base import SearchEngine
from search.elastic import ElasticSearchEngine
from search.tests.utils import SearcherMixin, TEST_INDEX_NAME
from search.utils import ValueRange, DateRange

from .mock_search_engine import MockSearchEngine


# Any class that inherits from TestCase will cause too-many-public-methods pylint error
# pylint: disable=too-many-public-methods
@override_settings(SEARCH_ENGINE="search.tests.mock_search_engine.MockSearchEngine")
@override_settings(ELASTIC_FIELD_MAPPINGS={"start_date": {"type": "date"}})
@override_settings(MOCK_SEARCH_BACKING_FILE=None)
class MockSearchTests(TestCase, SearcherMixin):
    """ Test operation of search activities """
    @property
    def _is_elastic(self):
        """ check search engine implementation, to manage cleanup differently """
        return isinstance(self.searcher, ElasticSearchEngine)

    def setUp(self):
        super(MockSearchTests, self).setUp()
        # ignore unexpected-keyword-arg; ES python client documents that it can be used
        # pylint: disable=unexpected-keyword-arg
        if self._is_elastic:
            _elasticsearch = Elasticsearch()
            # Make sure that we are fresh
            _elasticsearch.indices.delete(index=TEST_INDEX_NAME, ignore=[400, 404])

            config_body = {}
            # ignore unexpected-keyword-arg; ES python client documents that it can be used
            _elasticsearch.indices.create(index=TEST_INDEX_NAME, ignore=400, body=config_body)
        else:
            MockSearchEngine.destroy()
        self._searcher = None
        cache.clear()

    def tearDown(self):
        # ignore unexpected-keyword-arg; ES python client documents that it can be used
        # pylint: disable=unexpected-keyword-arg
        if self._is_elastic:
            _elasticsearch = Elasticsearch()
            # ignore unexpected-keyword-arg; ES python client documents that it can be used
            _elasticsearch.indices.delete(index=TEST_INDEX_NAME, ignore=[400, 404])
        else:
            MockSearchEngine.destroy()

        self._searcher = None
        super(MockSearchTests, self).tearDown()

    def test_factory_creator(self):
        """ Make sure that search object implements SearchEngine interface """
        self.assertTrue(isinstance(self.searcher, SearchEngine))

    def test_abstract_impl(self):
        """ Make sure that if one tries to use the abstract base, then operations yeild NotImplementedError s """
        abstract = SearchEngine("test_index_name")
        test_string = "A test string"
        with self.assertRaises(NotImplementedError):
            abstract.index("test_doc", [{"name": test_string}])
        with self.assertRaises(NotImplementedError):
            abstract.search(test_string)
        with self.assertRaises(NotImplementedError):
            abstract.remove("test_doc", ["test_id"])

    def test_find_all(self):
        """ Make sure that null search finds everything in the index """
        test_string = "A test string"
        self.searcher.index("test_doc", [{"name": test_string}])

        # search everything
        response = self.searcher.search(None)
        self.assertEqual(response["total"], 1)
        results = response["results"]
        self.assertEqual(results[0]["data"]["name"], test_string)

        self.searcher.index("not_test_doc", [{"value": test_string}])

        response = self.searcher.search(None)
        self.assertEqual(response["total"], 2)
        results = response["results"]
        test_0 = results[0]["data"] if "name" in results[0]["data"] else results[1]["data"]
        test_1 = results[1]["data"] if "name" in results[0]["data"] else results[0]["data"]
        self.assertEqual(test_0["name"], test_string)
        self.assertEqual(test_1["value"], test_string)

    def test_find_doctype(self):
        """ Make sure that searches for specific doc_type only return requested doc_type """
        test_string = "A test string"
        self.searcher.index("test_doc", [{"name": test_string}])

        # search by doc_type
        response = self.searcher.search(None, doc_type="test_doc")
        self.assertEqual(response["total"], 1)

        response = self.searcher.search(None, doc_type="not_test_doc")
        self.assertEqual(response["total"], 0)

        self.searcher.index("not_test_doc", [{"value": test_string}])

        response = self.searcher.search(None, doc_type="not_test_doc")
        self.assertEqual(response["total"], 1)

    def test_find_string(self):
        """ Find a string within the object "content" node """
        test_string = "A test string"
        self.searcher.index("test_doc", [{"content": {"name": test_string}}])

        # search string
        response = self.searcher.search(test_string)
        self.assertEqual(response["total"], 1)

        response = self.searcher.search_string(test_string)
        self.assertEqual(response["total"], 1)

        self.searcher.index("not_test_doc", [{"content": {"value": test_string}}])

        response = self.searcher.search_string(test_string)
        self.assertEqual(response["total"], 2)

        response = self.searcher.search_string(test_string, doc_type="test_doc")
        self.assertEqual(response["total"], 1)

        response = self.searcher.search_string("something else")
        self.assertEqual(response["total"], 0)

        self.searcher.index("test_doc", [{"content": {"deep": {"down": test_string}}}])
        response = self.searcher.search_string(test_string)
        self.assertEqual(response["total"], 3)

    def test_field(self):
        """ test matching on a field """
        test_string = "A test string"
        test_object = {
            "name": test_string,
            "tags": {
                "tag_one": "one",
                "tag_two": "two"
            },
            "fieldX": "valueY",
            "id": "12345"
        }
        self.searcher.index("test_doc", [test_object])

        # search tags
        response = self.searcher.search_fields({"tags.tag_one": "one"})
        self.assertEqual(response["total"], 1)

        # search id
        response = self.searcher.search_fields({"id": "12345"})
        self.assertEqual(response["total"], 1)

        # search id
        response = self.searcher.search_fields({"id": "54321"})
        self.assertEqual(response["total"], 0)

        # search tags
        response = self.searcher.search_fields({"tags.tag_one": "one", "tags.tag_two": "two"})
        self.assertEqual(response["total"], 1)

        response = self.searcher.search_fields({"fieldX": "valueY"})
        self.assertEqual(response["total"], 1)

        # search tags
        response = self.searcher.search_fields({"tags.tag_one": "one", "tags.tag_two": "not_two"})
        self.assertEqual(response["total"], 0)

    def test_search_string_and_field(self):
        """ test matching on both string and field value """
        test_object = {
            "content": {
                "name": "You may find me in a coffee shop",
            },
            "course": "A/B/C",
            "abc": "xyz",
        }
        self.searcher.index("test_doc", [test_object])

        response = self.searcher.search(query_string="find me")
        self.assertEqual(response["total"], 1)

        response = self.searcher.search_fields({"course": "A/B/C"})
        self.assertEqual(response["total"], 1)

        response = self.searcher.search(query_string="find me", field_dictionary={"course": "X/Y/Z"})
        self.assertEqual(response["total"], 0)

        response = self.searcher.search(query_string="find me", field_dictionary={"course": "A/B/C"})
        self.assertEqual(response["total"], 1)

        response = self.searcher.search_string("find me", field_dictionary={"course": "A/B/C"})
        self.assertEqual(response["total"], 1)

        response = self.searcher.search_fields({"course": "A/B/C"}, query_string="find me")
        self.assertEqual(response["total"], 1)

    def test_search_tags(self):
        """ test nested object tags """
        test_object = {
            "name": "John Lester",
            "course_id": "A/B/C",
            "abc": "xyz"
        }
        tags = {
            "color": "red",
            "shape": "square",
            "taste": "sour",
        }
        test_object["tags"] = tags
        self.searcher.index("test_doc", [test_object])

        response = self.searcher.search_fields({"tags.color": "red"})
        self.assertEqual(response["total"], 1)
        result = response["results"][0]["data"]
        self.assertEqual(result["tags"]["color"], "red")
        self.assertEqual(result["tags"]["shape"], "square")
        self.assertEqual(result["tags"]["taste"], "sour")

        response = self.searcher.search(field_dictionary={"tags.color": "red"})
        self.assertEqual(response["total"], 1)
        result = response["results"][0]["data"]
        self.assertEqual(result["tags"]["color"], "red")
        self.assertEqual(result["tags"]["shape"], "square")
        self.assertEqual(result["tags"]["taste"], "sour")

        response = self.searcher.search(field_dictionary={"tags.color": "blue"})
        self.assertEqual(response["total"], 0)

        response = self.searcher.search(field_dictionary={"tags.shape": "square"})
        self.assertEqual(response["total"], 1)
        result = response["results"][0]["data"]
        self.assertEqual(result["tags"]["color"], "red")
        self.assertEqual(result["tags"]["shape"], "square")
        self.assertEqual(result["tags"]["taste"], "sour")

        response = self.searcher.search(field_dictionary={"tags.shape": "round"})
        self.assertEqual(response["total"], 0)

        response = self.searcher.search(field_dictionary={"tags.shape": "square", "tags.color": "red"})
        self.assertEqual(response["total"], 1)
        result = response["results"][0]["data"]
        self.assertEqual(result["tags"]["color"], "red")
        self.assertEqual(result["tags"]["shape"], "square")
        self.assertEqual(result["tags"]["taste"], "sour")

        response = self._searcher.search(field_dictionary={"tags.shape": "square", "tags.color": "blue"})
        self.assertEqual(response["total"], 0)

        response = self._searcher.search(
            field_dictionary={"tags.shape": "square", "tags.color": "blue"}, use_field_match=True)
        self.assertEqual(response["total"], 0)

    def test_search_array(self):
        """ test nested object array """
        test_object1 = {
            "name": "John Lester",
            "course_id": "A/B/C",
            "array": ["a", "c", "x"]
        }
        test_object2 = {
            "name": "Anthony Rizzo",
            "course_id": "C/D/E",
            "array": ["a", "b", "c"]
        }
        self.searcher.index("test_doc", [test_object1])
        self.searcher.index("test_doc", [test_object2])

        response = self.searcher.search(field_dictionary={"array": "x"})
        self.assertEqual(response["total"], 1)
        self.assertEqual(response["results"][0]["data"], test_object1)

        response = self.searcher.search(field_dictionary={"array": "a"})
        self.assertEqual(response["total"], 2)
        self.assertIn(response["results"][0]["data"], [test_object1, test_object2])
        self.assertIn(response["results"][1]["data"], [test_object1, test_object2])

        response = self.searcher.search(field_dictionary={"array": "g"})
        self.assertEqual(response["total"], 0)

        response = self.searcher.search(field_dictionary={"array": "c"})
        self.assertEqual(response["total"], 2)

    def test_search_any(self):
        """ match any field value in a list """
        test_object1 = {
            "name": "John Lester",
            "course": "A/B/C"
        }
        test_object2 = {
            "name": "Anthony Rizzo",
            "course": "C/D/E"
        }
        self.searcher.index("test_doc", [test_object1])
        self.searcher.index("test_doc", [test_object2])

        response = self.searcher.search(field_dictionary={"course": ["x", "y"]})
        self.assertEqual(response["total"], 0)

        response = self.searcher.search(field_dictionary={"course": ["C/D/E"]})
        self.assertEqual(response["total"], 1)
        self.assertEqual(response["results"][0]["data"], test_object2)

        response = self.searcher.search(field_dictionary={"course": ["A/B/C", "C/D/E", "x"]})
        self.assertEqual(response["total"], 2)
        self.assertIn(response["results"][0]["data"], [test_object1, test_object2])
        self.assertIn(response["results"][1]["data"], [test_object1, test_object2])

    def test_extended_characters(self):
        """ Make sure that extended character searches work """
        test_string = u"قضايـا هامـة"
        self.searcher.index("test_doc", [{"content": {"name": test_string}}])

        # search string
        response = self.searcher.search(test_string)
        self.assertEqual(response["total"], 1)

        response = self.searcher.search_string(test_string)
        self.assertEqual(response["total"], 1)

        self.searcher.index("not_test_doc", [{"content": {"value": test_string}}])

        response = self.searcher.search_string(test_string)
        self.assertEqual(response["total"], 2)

    def test_delete_item(self):
        """ make sure that we can remove an item from the index """
        test_string = "This is a test of the emergency broadcast system"
        self.searcher.index("test_doc", [{"id": "FAKE_ID", "content": {"name": test_string}}])

        response = self.searcher.search(test_string)
        self.assertEqual(response["total"], 1)

        self.searcher.remove("test_doc", [response["results"][0]["data"]["id"]])
        response = self.searcher.search(test_string)
        self.assertEqual(response["total"], 0)

    def test_delete_item_slashes(self):
        """ make sure that we can remove an item from the index with complex id """
        test_string = "This is a test of the emergency broadcast system"
        self.searcher.index(
            "test_doc", [
                {
                    "id": "i4x://edX/DemoX/google-document/e3369ea4aa0749a7ba29c461d1c819a4",
                    "content": {"name": test_string}
                }
            ]
        )

        response = self.searcher.search(test_string)
        self.assertEqual(response["total"], 1)

        self.searcher.remove("test_doc", ["i4x://edX/DemoX/google-document/e3369ea4aa0749a7ba29c461d1c819a4"])
        response = self.searcher.search(test_string)
        self.assertEqual(response["total"], 0)

    def test_delete_item_not_present(self):
        """ make sure that we get no error removing an item that does not exist """
        self.searcher.remove("test_doc", ["TOTALLY_FAKE_ID"])

    def test_filter_items(self):
        """ Make sure that filters work """
        self.searcher.index("test_doc", [{"id": "FAKE_ID_1", "test_value": "1", "filter_field": "my_filter_value"}])
        self.searcher.index("test_doc", [{"id": "FAKE_ID_2", "test_value": "2"}])

        response = self.searcher.search(filter_dictionary={"filter_field": "my_filter_value"})
        self.assertEqual(response["total"], 2)

        response = self.searcher.search(field_dictionary={"filter_field": "my_filter_value"})
        self.assertEqual(response["total"], 1)

        self.searcher.index("test_doc", [{"id": "FAKE_ID_3", "test_value": "3", "filter_field": "not_my_filter_value"}])
        response = self.searcher.search(filter_dictionary={"filter_field": "my_filter_value"})
        self.assertEqual(response["total"], 2)

        response = self.searcher.search()
        self.assertEqual(response["total"], 3)

    def test_iterable_filters(self):
        """
        Make sure that iterable filters works
        """
        self.searcher.index("test_doc", [{"id": "FAKE_ID_1"}])
        self.searcher.index("test_doc", [{"id": "FAKE_ID_2", "filter_field": "orange"}])
        self.searcher.index("test_doc", [{"id": "FAKE_ID_3", "filter_field": ["orange", "blue"]}])

        response = self.searcher.search(filter_dictionary={"filter_field": "orange"})
        self.assertEqual(response["total"], 3)

        response = self.searcher.search(filter_dictionary={"filter_field": ["orange", "green"]})
        self.assertEqual(response["total"], 3)

        response = self.searcher.search(filter_dictionary={"filter_field": ["red", "green"]})
        self.assertEqual(response["total"], 1)

    def test_filter_where_null(self):
        """
        Make sure that filtering with `None` value finds only fields where item
        is not present or where explicitly None
        """
        self.searcher.index("test_doc", [{"id": "FAKE_ID_1", "test_value": "1", "filter_field": "my_filter_value"}])
        self.searcher.index("test_doc", [{"id": "FAKE_ID_2", "test_value": "2"}])
        self.searcher.index("test_doc", [{"id": "FAKE_ID_3", "test_value": "3", "filter_field": "not_my_filter_value"}])

        response = self.searcher.search(filter_dictionary={"filter_field": "my_filter_value"})
        self.assertEqual(response["total"], 2)

        response = self.searcher.search(filter_dictionary={"filter_field": None})
        self.assertEqual(response["total"], 1)

        self.searcher.index("test_doc", [{"id": "FAKE_ID_4", "test_value": "4", "filter_field": None}])
        response = self.searcher.search(filter_dictionary={"filter_field": None})
        self.assertEqual(response["total"], 2)

    def test_date_range(self):
        """ Make sure that date ranges can be searched """
        self.searcher.index("test_doc", [{"id": "FAKE_ID_1", "test_value": "1", "start_date": datetime(2010, 1, 1)}])
        self.searcher.index("test_doc", [{"id": "FAKE_ID_2", "test_value": "2", "start_date": datetime(2100, 1, 1)}])

        response = self.searcher.search()
        self.assertEqual(response["total"], 2)

        response = self.searcher.search(field_dictionary={"start_date": DateRange(None, datetime.utcnow())})
        self.assertEqual(response["total"], 1)

        response = self.searcher.search(field_dictionary={"start_date": DateRange(datetime(2099, 1, 1), None)})
        self.assertEqual(response["total"], 1)

        response = self.searcher.search(filter_dictionary={"start_date": DateRange(None, datetime.utcnow())})
        self.assertEqual(response["total"], 1)

        response = self.searcher.search(filter_dictionary={"start_date": DateRange(datetime(2099, 1, 1), None)})
        self.assertEqual(response["total"], 1)

    def test_numeric_range(self):
        """ Make sure that numeric ranges can be searched with both field and filter queries """
        self.searcher.index("test_doc", [{"id": "FAKE_ID_1", "test_value": "1", "age": 20}])
        self.searcher.index("test_doc", [{"id": "FAKE_ID_2", "test_value": "2", "age": 30}])
        self.searcher.index("test_doc", [{"id": "FAKE_ID_3", "test_value": "3", "age": 40}])

        def test_age_range_field(begin, end, expect):
            """ repeated operations consolidated for tests """
            response = self.searcher.search(field_dictionary={"age": ValueRange(begin, end)})
            self.assertEqual(response["total"], expect)

        def test_age_range_filter(begin, end, expect):
            """ repeated operations consolidated for tests """
            response = self.searcher.search(filter_dictionary={"age": ValueRange(begin, end)})
            self.assertEqual(response["total"], expect)

        response = self.searcher.search()
        self.assertEqual(response["total"], 3)

        test_age_range_field(19, 29, 1)
        test_age_range_field(19, 39, 2)
        test_age_range_field(19, 49, 3)
        test_age_range_field(29, 49, 2)
        test_age_range_field(39, 49, 1)
        test_age_range_field(None, 29, 1)
        test_age_range_field(39, None, 1)

        test_age_range_filter(19, 29, 1)
        test_age_range_filter(19, 39, 2)
        test_age_range_filter(19, 49, 3)
        test_age_range_filter(29, 49, 2)
        test_age_range_filter(39, 49, 1)
        test_age_range_filter(None, 29, 1)
        test_age_range_filter(39, None, 1)

        self.searcher.index("test_doc", [{"id": "FAKE_ID_4", "test_value": "4", "age": 50}])

        test_age_range_field(19, 29, 1)
        test_age_range_field(19, 39, 2)
        test_age_range_field(19, 49, 3)
        test_age_range_field(29, 49, 2)
        test_age_range_field(39, 49, 1)
        test_age_range_field(None, 29, 1)
        test_age_range_field(39, None, 2)

        test_age_range_filter(19, 29, 1)
        test_age_range_filter(19, 39, 2)
        test_age_range_filter(19, 49, 3)
        test_age_range_filter(29, 49, 2)
        test_age_range_filter(39, 49, 1)
        test_age_range_filter(None, 29, 1)
        test_age_range_filter(39, None, 2)

        self.searcher.index("test_doc", [{"id": "FAKE_ID_5", "test_value": "5", "not_age": 50}])
        test_age_range_filter(19, 29, 2)
        test_age_range_filter(19, 39, 3)
        test_age_range_filter(19, 49, 4)
        test_age_range_filter(29, 49, 3)
        test_age_range_filter(39, 49, 2)
        test_age_range_filter(None, 29, 2)
        test_age_range_filter(39, None, 3)

    def test_range_filter(self):
        """ Make sure that ranges can be used in field_dictionary and filter_dictionary """
        self.searcher.index("test_doc", [{"id": "FAKE_ID_1", "test_value": "1", "age": 20}])
        self.searcher.index("test_doc", [{"id": "FAKE_ID_2", "test_value": "2", "age": 30}])
        self.searcher.index("test_doc", [{"id": "FAKE_ID_3", "test_value": "3", "not_age": 40}])

        response = self.searcher.search()
        self.assertEqual(response["total"], 3)

        response = self.searcher.search(field_dictionary={"age": ValueRange(19, 29)})
        self.assertEqual(response["total"], 1)

        response = self.searcher.search(filter_dictionary={"age": ValueRange(19, 29)})
        self.assertEqual(response["total"], 2)

    def test_pagination(self):
        """ Test paging operation """
        self.searcher.index(
            "test_doc",
            [
                {
                    "course": "ABC",
                    "id": "FAKE_ID_1",
                    "content": {
                        "text": "Little Darling Little Darling Little Darling, it's been a long long lonely winter"
                    }
                }
            ]
        )
        self.searcher.index(
            "test_doc",
            [
                {
                    "course": "ABC",
                    "id": "FAKE_ID_2",
                    "content": {
                        "text": "Little Darling Little Darling, it's been a year since you've been gone"
                    }
                }
            ]
        )
        self.searcher.index(
            "test_doc",
            [
                {
                    "course": "XYZ",
                    "id": "FAKE_ID_3",
                    "content": {
                        "text": "Little Darling, it's been a long long lonely winter"
                    }
                }
            ]
        )

        response = self.searcher.search(query_string="Little Darling")
        self.assertEqual(response["total"], 3)
        self.assertEqual(len(response["results"]), 3)

        response = self.searcher.search(query_string="Little Darling", size=1)
        self.assertEqual(response["total"], 3)
        self.assertEqual(len(response["results"]), 1)
        result_ids = [r["data"]["id"] for r in response["results"]]
        self.assertIn("FAKE_ID_1", result_ids)

        response = self.searcher.search(query_string="Little Darling", size=1, from_=0)
        self.assertEqual(response["total"], 3)
        result_ids = [r["data"]["id"] for r in response["results"]]
        self.assertEqual(result_ids, ["FAKE_ID_1"])

        response = self.searcher.search(query_string="Little Darling", size=1, from_=1)
        self.assertEqual(response["total"], 3)
        result_ids = [r["data"]["id"] for r in response["results"]]
        self.assertEqual(result_ids, ["FAKE_ID_2"])

        response = self.searcher.search(query_string="Little Darling", size=1, from_=2)
        self.assertEqual(response["total"], 3)
        result_ids = [r["data"]["id"] for r in response["results"]]
        self.assertEqual(result_ids, ["FAKE_ID_3"])

        response = self.searcher.search(query_string="Little Darling", size=2)
        self.assertEqual(response["total"], 3)
        result_ids = [r["data"]["id"] for r in response["results"]]
        self.assertEqual(result_ids, ["FAKE_ID_1", "FAKE_ID_2"])

        response = self.searcher.search(query_string="Little Darling", size=2, from_=0)
        self.assertEqual(response["total"], 3)
        result_ids = [r["data"]["id"] for r in response["results"]]
        self.assertEqual(result_ids, ["FAKE_ID_1", "FAKE_ID_2"])

        response = self.searcher.search(query_string="Little Darling", size=2, from_=2)
        self.assertEqual(response["total"], 3)
        result_ids = [r["data"]["id"] for r in response["results"]]
        self.assertEqual(result_ids, ["FAKE_ID_3"])

    def test_exclude_ids(self):
        """ Test that ids that would normally be present in the resultset will not be present if in the exclude list """
        self.searcher.index("test_doc", [{"course": "ABC", "id": "FAKE_ID_1"}])
        self.searcher.index("test_doc", [{"course": "ABC", "id": "FAKE_ID_2"}])
        self.searcher.index("test_doc", [{"course": "ABC", "id": "FAKE_ID_3"}])
        self.searcher.index("test_doc", [{"course": "XYZ", "id": "FAKE_ID_11"}])
        self.searcher.index("test_doc", [{"course": "XYZ", "id": "FAKE_ID_12"}])
        self.searcher.index("test_doc", [{"course": "XYZ", "id": "FAKE_ID_13"}])

        response = self.searcher.search()
        self.assertEqual(response["total"], 6)
        result_ids = [r["data"]["id"] for r in response["results"]]
        self.assertIn("FAKE_ID_1", result_ids)
        self.assertIn("FAKE_ID_2", result_ids)
        self.assertIn("FAKE_ID_3", result_ids)
        self.assertIn("FAKE_ID_11", result_ids)
        self.assertIn("FAKE_ID_12", result_ids)
        self.assertIn("FAKE_ID_13", result_ids)

        response = self.searcher.search(exclude_ids=["FAKE_ID_1", "FAKE_ID_2", "FAKE_ID_11", "FAKE_ID_12"])
        self.assertEqual(response["total"], 2)
        result_ids = [r["data"]["id"] for r in response["results"]]
        self.assertNotIn("FAKE_ID_1", result_ids)
        self.assertNotIn("FAKE_ID_2", result_ids)
        self.assertIn("FAKE_ID_3", result_ids)
        self.assertNotIn("FAKE_ID_11", result_ids)
        self.assertNotIn("FAKE_ID_12", result_ids)
        self.assertIn("FAKE_ID_13", result_ids)

        response = self.searcher.search(field_dictionary={"course": "ABC"})
        self.assertEqual(response["total"], 3)
        result_ids = [r["data"]["id"] for r in response["results"]]
        self.assertIn("FAKE_ID_1", result_ids)
        self.assertIn("FAKE_ID_2", result_ids)
        self.assertIn("FAKE_ID_3", result_ids)
        self.assertNotIn("FAKE_ID_11", result_ids)
        self.assertNotIn("FAKE_ID_12", result_ids)
        self.assertNotIn("FAKE_ID_13", result_ids)

        response = self.searcher.search(field_dictionary={"course": "ABC"}, exclude_ids=["FAKE_ID_3", "FAKE_ID_2"])
        self.assertEqual(response["total"], 1)
        result_ids = [r["data"]["id"] for r in response["results"]]
        self.assertIn("FAKE_ID_1", result_ids)
        self.assertNotIn("FAKE_ID_2", result_ids)
        self.assertNotIn("FAKE_ID_3", result_ids)
        self.assertNotIn("FAKE_ID_11", result_ids)
        self.assertNotIn("FAKE_ID_12", result_ids)
        self.assertNotIn("FAKE_ID_13", result_ids)

    def test_exclude_filter_single(self):
        """ Test that single entries present in the exclude filter are filtered out """
        self.searcher.index("test_doc", [{"course": "ABC", "org": "edX", "id": "FAKE_ID_1"}])
        self.searcher.index("test_doc", [{"course": "XYZ", "org": "edX", "id": "FAKE_ID_2"}])
        self.searcher.index("test_doc", [{"course": "LMN", "org": "MITX", "id": "FAKE_ID_3"}])

        response = self.searcher.search()
        self.assertEqual(response["total"], 3)
        result_ids = [r["data"]["id"] for r in response["results"]]
        self.assertIn("FAKE_ID_1", result_ids)
        self.assertIn("FAKE_ID_2", result_ids)
        self.assertIn("FAKE_ID_3", result_ids)

        response = self.searcher.search(exclude_dictionary={"org": "MITX"})
        self.assertEqual(response["total"], 2)
        result_ids = [r["data"]["id"] for r in response["results"]]
        self.assertIn("FAKE_ID_1", result_ids)
        self.assertIn("FAKE_ID_2", result_ids)
        self.assertNotIn("FAKE_ID_3", result_ids)

        response = self.searcher.search(exclude_dictionary={"org": "edX"})
        self.assertEqual(response["total"], 1)
        result_ids = [r["data"]["id"] for r in response["results"]]
        self.assertNotIn("FAKE_ID_1", result_ids)
        self.assertNotIn("FAKE_ID_2", result_ids)
        self.assertIn("FAKE_ID_3", result_ids)

        # Combo excludes
        response = self.searcher.search(exclude_dictionary={"org": "MITX", "course": "ABC"})
        self.assertEqual(response["total"], 1)
        result_ids = [r["data"]["id"] for r in response["results"]]
        self.assertNotIn("FAKE_ID_1", result_ids)
        self.assertIn("FAKE_ID_2", result_ids)
        self.assertNotIn("FAKE_ID_3", result_ids)

        # ignore with include
        response = self.searcher.search(field_dictionary={"course": "ABC"}, exclude_dictionary={"org": "MITX"})
        self.assertEqual(response["total"], 1)
        result_ids = [r["data"]["id"] for r in response["results"]]
        self.assertIn("FAKE_ID_1", result_ids)
        self.assertNotIn("FAKE_ID_2", result_ids)
        self.assertNotIn("FAKE_ID_3", result_ids)

    def test_exclude_filter_multiple(self):
        """ Test that multiple entries present in the exclude filter are filtered out """
        self.searcher.index("test_doc", [{"course": "ABC", "org": "edX", "id": "FAKE_ID_1"}])
        self.searcher.index("test_doc", [{"course": "XYZ", "org": "edX", "id": "FAKE_ID_2"}])
        self.searcher.index("test_doc", [{"course": "DEF", "org": "MITX", "id": "FAKE_ID_3"}])
        self.searcher.index("test_doc", [{"course": "GHI", "org": "HarvardX", "id": "FAKE_ID_4"}])
        self.searcher.index("test_doc", [{"course": "LMN", "org": "edX", "id": "FAKE_ID_5"}])

        response = self.searcher.search()
        self.assertEqual(response["total"], 5)
        result_ids = [r["data"]["id"] for r in response["results"]]
        self.assertIn("FAKE_ID_1", result_ids)
        self.assertIn("FAKE_ID_2", result_ids)
        self.assertIn("FAKE_ID_3", result_ids)
        self.assertIn("FAKE_ID_4", result_ids)
        self.assertIn("FAKE_ID_5", result_ids)

        response = self.searcher.search(exclude_dictionary={"org": ["MITX", "HarvardX"]})
        self.assertEqual(response["total"], 3)
        result_ids = [r["data"]["id"] for r in response["results"]]
        self.assertIn("FAKE_ID_1", result_ids)
        self.assertIn("FAKE_ID_2", result_ids)
        self.assertNotIn("FAKE_ID_3", result_ids)
        self.assertNotIn("FAKE_ID_4", result_ids)
        self.assertIn("FAKE_ID_5", result_ids)

        response = self.searcher.search(exclude_dictionary={"org": "edX"})
        self.assertEqual(response["total"], 2)
        result_ids = [r["data"]["id"] for r in response["results"]]
        self.assertNotIn("FAKE_ID_1", result_ids)
        self.assertNotIn("FAKE_ID_2", result_ids)
        self.assertIn("FAKE_ID_3", result_ids)
        self.assertIn("FAKE_ID_4", result_ids)
        self.assertNotIn("FAKE_ID_5", result_ids)

        # Combo excludes
        response = self.searcher.search(
            exclude_dictionary={"org": ["MITX", "HarvardX"], "course": ["XYZ", "LMN", "DEF"]}
        )
        self.assertEqual(response["total"], 1)
        result_ids = [r["data"]["id"] for r in response["results"]]
        self.assertIn("FAKE_ID_1", result_ids)
        self.assertNotIn("FAKE_ID_2", result_ids)
        self.assertNotIn("FAKE_ID_3", result_ids)
        self.assertNotIn("FAKE_ID_4", result_ids)
        self.assertNotIn("FAKE_ID_5", result_ids)

        # ignore with include
        response = self.searcher.search(
            field_dictionary={"course": ["XYZ", "LMN", "DEF"]},
            exclude_dictionary={"org": ["MITX", "HarvardX"]}
        )
        self.assertEqual(response["total"], 2)
        result_ids = [r["data"]["id"] for r in response["results"]]
        self.assertNotIn("FAKE_ID_1", result_ids)
        self.assertIn("FAKE_ID_2", result_ids)
        self.assertNotIn("FAKE_ID_3", result_ids)
        self.assertNotIn("FAKE_ID_4", result_ids)
        self.assertIn("FAKE_ID_5", result_ids)

    def test_exclude_filter_empty(self):
        """ Test that search works when exclude filter is an empty list """
        self.searcher.index("test_doc", [{"course": "ABC", "org": "edX", "id": "FAKE_ID_1"}])
        self.searcher.index("test_doc", [{"course": "XYZ", "org": "edX", "id": "FAKE_ID_2"}])
        self.searcher.index("test_doc", [{"course": "DEF", "org": "MITX", "id": "FAKE_ID_3"}])
        self.searcher.index("test_doc", [{"course": "GHI", "org": "HarvardX", "id": "FAKE_ID_4"}])
        self.searcher.index("test_doc", [{"course": "LMN", "org": "edX", "id": "FAKE_ID_5"}])

        response = self.searcher.search(exclude_dictionary={"org": []})
        self.assertEqual(response["total"], 5)
        result_ids = [r["data"]["id"] for r in response["results"]]
        self.assertIn("FAKE_ID_1", result_ids)
        self.assertIn("FAKE_ID_2", result_ids)
        self.assertIn("FAKE_ID_3", result_ids)
        self.assertIn("FAKE_ID_4", result_ids)
        self.assertIn("FAKE_ID_5", result_ids)

        response = self.searcher.search(
            field_dictionary={"course": ["XYZ", "LMN", "DEF"]},
            exclude_dictionary={"org": []}
        )
        self.assertEqual(response["total"], 3)
        result_ids = [r["data"]["id"] for r in response["results"]]
        self.assertNotIn("FAKE_ID_1", result_ids)
        self.assertIn("FAKE_ID_2", result_ids)
        self.assertIn("FAKE_ID_3", result_ids)
        self.assertNotIn("FAKE_ID_4", result_ids)
        self.assertIn("FAKE_ID_5", result_ids)

    def _index_for_facets(self):
        """ Prepare index for facet tests """
        self.searcher.index("test_doc", [{"id": "FAKE_ID_1", "subject": "mathematics", "org": "edX"}])
        self.searcher.index("test_doc", [{"id": "FAKE_ID_2", "subject": "mathematics", "org": "MIT"}])
        self.searcher.index("test_doc", [{"id": "FAKE_ID_3", "subject": "physics", "org": "MIT"}])
        self.searcher.index("test_doc", [{"id": "FAKE_ID_4", "subject": "history", "org": "Harvard"}])
        self.searcher.index("test_doc", [{"id": "FAKE_ID_5", "subject": "mathematics", "org": "Harvard"}])
        self.searcher.index("test_doc", [{"id": "FAKE_ID_6", "subject": "physics", "org": "Harvard"}])
        self.searcher.index("test_doc", [{"id": "FAKE_ID_7", "no_subject": "not_a_subject", "org": "Harvard"}])

    def test_faceted_search(self):
        """ Test that faceting works well """
        self._index_for_facets()

        response = self.searcher.search()
        self.assertEqual(response["total"], 7)
        self.assertNotIn("facets", response)

        facet_terms = {
            "subject": {},
            "org": {}
        }
        response = self.searcher.search(facet_terms=facet_terms)
        self.assertEqual(response["total"], 7)
        self.assertIn("facets", response)
        facet_results = response["facets"]

        self.assertEqual(facet_results["subject"]["total"], 6)
        subject_term_counts = facet_results["subject"]["terms"]
        self.assertEqual(subject_term_counts["mathematics"], 3)
        self.assertEqual(subject_term_counts["history"], 1)
        self.assertEqual(subject_term_counts["physics"], 2)

        self.assertEqual(facet_results["org"]["total"], 7)
        org_term_counts = facet_results["org"]["terms"]
        self.assertEqual(org_term_counts["edX"], 1)
        self.assertEqual(org_term_counts["MIT"], 2)
        self.assertEqual(org_term_counts["Harvard"], 4)

    def test_filtered_faceted_search(self):
        """ Test that faceting works well alongside filtered results """
        self._index_for_facets()

        facet_terms = {
            "subject": {},
            "org": {}
        }
        response = self.searcher.search(field_dictionary={"org": "Harvard"}, facet_terms=facet_terms)
        self.assertEqual(response["total"], 4)
        self.assertIn("facets", response)
        facet_results = response["facets"]

        self.assertEqual(facet_results["subject"]["total"], 3)
        subject_term_counts = facet_results["subject"]["terms"]
        self.assertEqual(subject_term_counts["mathematics"], 1)
        self.assertEqual(subject_term_counts["history"], 1)
        self.assertEqual(subject_term_counts["physics"], 1)

        self.assertEqual(facet_results["org"]["total"], 4)
        org_term_counts = facet_results["org"]["terms"]
        self.assertNotIn("edX", org_term_counts)
        self.assertNotIn("MIT", org_term_counts)
        self.assertEqual(org_term_counts["Harvard"], 4)

        response = self.searcher.search(field_dictionary={"subject": ["physics", "history"]}, facet_terms=facet_terms)
        self.assertEqual(response["total"], 3)
        self.assertIn("facets", response)
        facet_results = response["facets"]

        self.assertEqual(facet_results["subject"]["total"], 3)
        subject_term_counts = facet_results["subject"]["terms"]
        self.assertNotIn("mathematics", subject_term_counts)
        self.assertEqual(subject_term_counts["history"], 1)
        self.assertEqual(subject_term_counts["physics"], 2)

        self.assertEqual(facet_results["org"]["total"], 3)
        org_term_counts = facet_results["org"]["terms"]
        self.assertNotIn("edX", org_term_counts)
        self.assertEqual(org_term_counts["MIT"], 1)
        self.assertEqual(org_term_counts["Harvard"], 2)



""" High-level view tests"""
from __future__ import absolute_import
from datetime import datetime
import ddt

from django.core.urlresolvers import Resolver404, resolve
from django.test import TestCase
from django.test.utils import override_settings
from mock import patch, call

import six
from search.search_engine_base import SearchEngine
from search.tests.mock_search_engine import MockSearchEngine
from search.tests.tests import TEST_INDEX_NAME
from search.tests.utils import post_discovery_request, post_request, SearcherMixin


# Any class that inherits from TestCase will cause too-many-public-methods pylint error
# pylint: disable=too-many-public-methods
@override_settings(SEARCH_ENGINE="search.tests.mock_search_engine.MockSearchEngine")
@override_settings(ELASTIC_FIELD_MAPPINGS={"start_date": {"type": "date"}})
@override_settings(COURSEWARE_INDEX_NAME=TEST_INDEX_NAME)
class MockSearchUrlTest(TestCase, SearcherMixin):
    """
    Make sure that requests to the url get routed to the correct view handler
    """

    def _reset_mocked_tracker(self):
        """ reset mocked tracker and clear logged emits """
        self.mock_tracker.reset_mock()

    def setUp(self):
        super(MockSearchUrlTest, self).setUp()
        MockSearchEngine.destroy()
        self._searcher = None
        patcher = patch('search.views.track')
        self.mock_tracker = patcher.start()
        self.addCleanup(patcher.stop)

    def tearDown(self):
        MockSearchEngine.destroy()
        self._searcher = None
        super(MockSearchUrlTest, self).tearDown()

    def assert_no_events_were_emitted(self):
        """Ensures no events were emitted since the last event related assertion"""
        self.assertFalse(self.mock_tracker.emit.called)  # pylint: disable=maybe-no-member

    def assert_search_initiated_event(self, search_term, size, page):
        """Ensures an search initiated event was emitted"""
        initiated_search_call = self.mock_tracker.emit.mock_calls[0]  # pylint: disable=maybe-no-member
        expected_result = call('edx.course.search.initiated', {
            "search_term": six.text_type(search_term),
            "page_size": size,
            "page_number": page,
        })
        self.assertEqual(expected_result, initiated_search_call)

    def assert_results_returned_event(self, search_term, size, page, total):
        """Ensures an results returned event was emitted"""
        returned_results_call = self.mock_tracker.emit.mock_calls[1]  # pylint: disable=maybe-no-member
        expected_result = call('edx.course.search.results_displayed', {
            "search_term": six.text_type(search_term),
            "page_size": size,
            "page_number": page,
            "results_count": total,
        })
        self.assertEqual(expected_result, returned_results_call)

    def assert_initiated_return_events(self, search_term, size, page, total):
        """Asserts search initiated and results returned events were emitted"""
        self.assertEqual(self.mock_tracker.emit.call_count, 2)  # pylint: disable=maybe-no-member
        self.assert_search_initiated_event(search_term, size, page)
        self.assert_results_returned_event(search_term, size, page, total)

    def test_url_resolution(self):
        """ make sure that the url is resolved as expected """
        resolver = resolve('/')
        self.assertEqual(resolver.view_name, 'do_search')

        with self.assertRaises(Resolver404):
            resolver = resolve('/blah')

        resolver = resolve('/edX/DemoX/Demo_Course')
        self.assertEqual(resolver.view_name, 'do_search')
        self.assertEqual(resolver.kwargs['course_id'], 'edX/DemoX/Demo_Course')

    def test_search_from_url(self):
        """ test searching using the url """
        self.searcher.index(
            "courseware_content",
            [
                {
                    "id": "FAKE_ID_1",
                    "content": {
                        "text": "Little Darling, it's been a long long lonely winter"
                    },
                    "test_date": datetime(2015, 1, 1),
                    "test_string": "ABC, It's easy as 123"
                }
            ]
        )
        self.searcher.index(
            "courseware_content",
            [
                {
                    "id": "FAKE_ID_2",
                    "content": {
                        "text": "Little Darling, it's been a year since sun been gone"
                    }
                }
            ]
        )
        self.searcher.index("courseware_content", [{"id": "FAKE_ID_3", "content": {"text": "Here comes the sun"}}])

        # Test no events called  yet after setup
        self.assert_no_events_were_emitted()
        self._reset_mocked_tracker()

        code, results = post_request({"search_string": "sun"})
        self.assertTrue(code < 300 and code > 199)
        self.assertEqual(results["total"], 2)
        result_ids = [r["data"]["id"] for r in results["results"]]
        self.assertTrue("FAKE_ID_3" in result_ids and "FAKE_ID_2" in result_ids)

        # Test initiate search and return results were called - and clear mocked tracker
        self.assert_initiated_return_events("sun", 20, 0, 2)
        self._reset_mocked_tracker()

        code, results = post_request({"search_string": "Darling"})
        self.assertTrue(code < 300 and code > 199)
        self.assertEqual(results["total"], 2)
        result_ids = [r["data"]["id"] for r in results["results"]]
        self.assertTrue("FAKE_ID_1" in result_ids and "FAKE_ID_2" in result_ids)

        # Test initiate search and return results were called - and clear mocked tracker
        self.assert_initiated_return_events("Darling", 20, 0, 2)
        self._reset_mocked_tracker()

        code, results = post_request({"search_string": "winter"})
        self.assertTrue(code < 300 and code > 199)
        self.assertEqual(results["total"], 1)
        result_ids = [r["data"]["id"] for r in results["results"]]
        self.assertTrue("FAKE_ID_1" in result_ids and "FAKE_ID_2" not in result_ids)

        # Test initiate search and return results were called - and clear mocked tracker
        self.assert_initiated_return_events("winter", 20, 0, 1)
        self._reset_mocked_tracker()

        self.assertTrue(results["results"][0]["data"]["test_date"], datetime(2015, 1, 1).isoformat())
        self.assertTrue(results["results"][0]["data"]["test_string"], "ABC, It's easy as 123")

    def test_course_search_url(self):
        """ test searching using the course url """
        self.searcher.index(
            "courseware_content",
            [
                {
                    "course": "ABC/DEF/GHI",
                    "id": "FAKE_ID_1",
                    "content": {
                        "text": "Little Darling, it's been a long long lonely winter"
                    }
                }
            ]
        )
        self.searcher.index(
            "courseware_content",
            [
                {
                    "course": "ABC/DEF/GHI",
                    "id": "FAKE_ID_2",
                    "content": {
                        "text": "Little Darling, it's been a year since you've been gone"
                    }
                }
            ]
        )
        self.searcher.index(
            "courseware_content",
            [
                {
                    "course": "LMN/OPQ/RST",
                    "id": "FAKE_ID_3",
                    "content": {
                        "text": "Little Darling, it's been a long long lonely winter"
                    }
                }
            ]
        )

        # Test no events called  yet after setup
        self.assert_no_events_were_emitted()
        self._reset_mocked_tracker()

        code, results = post_request({"search_string": "Little Darling"})
        self.assertTrue(code < 300 and code > 199)
        self.assertEqual(results["total"], 3)

        # Test initiate search and return results were called - and clear mocked tracker
        self.assert_initiated_return_events("Little Darling", 20, 0, 3)
        self._reset_mocked_tracker()

        code, results = post_request({"search_string": "Darling"}, "ABC/DEF/GHI")
        self.assertTrue(code < 300 and code > 199)
        self.assertEqual(results["total"], 2)
        result_ids = [r["data"]["id"] for r in results["results"]]
        self.assertTrue("FAKE_ID_1" in result_ids and "FAKE_ID_2" in result_ids)

        # Test initiate search and return results were called - and clear mocked tracker
        self.assert_initiated_return_events("Darling", 20, 0, 2)
        self._reset_mocked_tracker()

        code, results = post_request({"search_string": "winter"}, "ABC/DEF/GHI")
        self.assertTrue(code < 300 and code > 199)
        self.assertEqual(results["total"], 1)
        result_ids = [r["data"]["id"] for r in results["results"]]
        self.assertTrue("FAKE_ID_1" in result_ids and "FAKE_ID_2" not in result_ids and "FAKE_ID_3" not in result_ids)

        # Test initiate search and return results were called - and clear mocked tracker
        self.assert_initiated_return_events("winter", 20, 0, 1)
        self._reset_mocked_tracker()

        code, results = post_request({"search_string": "winter"}, "LMN/OPQ/RST")
        self.assertTrue(code < 300 and code > 199)
        self.assertEqual(results["total"], 1)
        result_ids = [r["data"]["id"] for r in results["results"]]
        self.assertTrue("FAKE_ID_1" not in result_ids and "FAKE_ID_2" not in result_ids and "FAKE_ID_3" in result_ids)

        # Test initiate search and return results were called - and clear mocked tracker
        self.assert_initiated_return_events("winter", 20, 0, 1)
        self._reset_mocked_tracker()

    def test_empty_search_string(self):
        """ test when search string is provided as empty or null (None) """
        code, results = post_request({"search_string": ""})
        self.assertGreater(code, 499)
        self.assertEqual(results["error"], "No search term provided for search")

        code, results = post_request({"no_search_string_provided": ""})
        self.assertGreater(code, 499)
        self.assertEqual(results["error"], "No search term provided for search")

    # pylint: disable=too-many-statements,wrong-assert-type
    def test_pagination(self):
        """ test searching using the course url """
        self.searcher.index(
            "courseware_content",
            [
                {
                    "course": "ABC",
                    "id": "FAKE_ID_1",
                    "content": {
                        "text": "Little Darling Little Darling Little Darling, it's been a long long lonely winter"
                    }
                }
            ]
        )
        self.searcher.index(
            "courseware_content",
            [
                {
                    "course": "ABC",
                    "id": "FAKE_ID_2",
                    "content": {
                        "text": "Little Darling Little Darling, it's been a year since you've been gone"
                    }
                }
            ]
        )
        self.searcher.index(
            "courseware_content",
            [
                {
                    "course": "XYZ",
                    "id": "FAKE_ID_3",
                    "content": {
                        "text": "Little Darling, it's been a long long lonely winter"
                    }
                }
            ]
        )

        # Test no events called  yet after setup
        self.assert_no_events_were_emitted()
        self._reset_mocked_tracker()

        code, results = post_request({"search_string": "Little Darling"})
        self.assertTrue(199 < code < 300)
        self.assertEqual(results["total"], 3)
        self.assertEqual(len(results["results"]), 3)

        # Test initiate search and return results were called - and clear mocked tracker
        self.assert_initiated_return_events("Little Darling", 20, 0, 3)
        self._reset_mocked_tracker()

        code, results = post_request({"search_string": "Little Darling", "page_size": 1})
        self.assertTrue(199 < code < 300)
        self.assertEqual(results["total"], 3)
        result_ids = [r["data"]["id"] for r in results["results"]]
        self.assertEqual(result_ids, ["FAKE_ID_1"])

        # Test initiate search and return results were called - and clear mocked tracker
        self.assert_initiated_return_events("Little Darling", 1, 0, 3)
        self._reset_mocked_tracker()

        code, results = post_request({"search_string": "Little Darling", "page_size": 1, "page_index": 0})
        self.assertTrue(199 < code < 300)
        self.assertEqual(results["total"], 3)
        result_ids = [r["data"]["id"] for r in results["results"]]
        self.assertEqual(result_ids, ["FAKE_ID_1"])

        # Test initiate search and return results were called - and clear mocked tracker
        self.assert_initiated_return_events("Little Darling", 1, 0, 3)
        self._reset_mocked_tracker()

        code, results = post_request({"search_string": "Little Darling", "page_size": 1, "page_index": 1})
        self.assertTrue(199 < code < 300)
        self.assertEqual(results["total"], 3)
        result_ids = [r["data"]["id"] for r in results["results"]]
        self.assertEqual(result_ids, ["FAKE_ID_2"])

        # Test initiate search and return results were called - and clear mocked tracker
        self.assert_initiated_return_events("Little Darling", 1, 1, 3)
        self._reset_mocked_tracker()

        code, results = post_request({"search_string": "Little Darling", "page_size": 1, "page_index": 2})
        self.assertTrue(199 < code < 300)
        self.assertEqual(results["total"], 3)
        result_ids = [r["data"]["id"] for r in results["results"]]
        self.assertEqual(result_ids, ["FAKE_ID_3"])

        # Test initiate search and return results were called - and clear mocked tracker
        self.assert_initiated_return_events("Little Darling", 1, 2, 3)
        self._reset_mocked_tracker()

        code, results = post_request({"search_string": "Little Darling", "page_size": 2})
        self.assertTrue(199 < code < 300)
        self.assertEqual(results["total"], 3)
        result_ids = [r["data"]["id"] for r in results["results"]]
        self.assertEqual(result_ids, ["FAKE_ID_1", "FAKE_ID_2"])

        # Test initiate search and return results were called - and clear mocked tracker
        self.assert_initiated_return_events("Little Darling", 2, 0, 3)
        self._reset_mocked_tracker()

        code, results = post_request({"search_string": "Little Darling", "page_size": 2, "page_index": 0})
        self.assertTrue(199 < code < 300)
        self.assertEqual(results["total"], 3)
        result_ids = [r["data"]["id"] for r in results["results"]]
        self.assertEqual(result_ids, ["FAKE_ID_1", "FAKE_ID_2"])

        # Test initiate search and return results were called - and clear mocked tracker
        self.assert_initiated_return_events("Little Darling", 2, 0, 3)
        self._reset_mocked_tracker()

        code, results = post_request({"search_string": "Little Darling", "page_size": 2, "page_index": 1})
        self.assertTrue(199 < code < 300)
        self.assertEqual(results["total"], 3)
        result_ids = [r["data"]["id"] for r in results["results"]]
        self.assertEqual(result_ids, ["FAKE_ID_3"])

        # Test initiate search and return results were called - and clear mocked tracker
        self.assert_initiated_return_events("Little Darling", 2, 1, 3)
        self._reset_mocked_tracker()

    def test_page_size_too_large(self):
        """ test searching with too-large page_size """
        self.searcher.index(
            "test_doc",
            [
                {
                    "course": "ABC/DEF/GHI",
                    "id": "FAKE_ID_1",
                    "content": {
                        "text": "Little Darling, it's been a long long lonely winter"
                    }
                }
            ]
        )

        code, results = post_request({"search_string": "Little Darling", "page_size": 101})
        self.assertEqual(code, 500)
        self.assertTrue("error" in results)


@override_settings(SEARCH_ENGINE="search.tests.utils.ErroringSearchEngine")
@override_settings(ELASTIC_FIELD_MAPPINGS={"start_date": {"type": "date"}})
@override_settings(COURSEWARE_INDEX_NAME=TEST_INDEX_NAME)
class BadSearchTest(TestCase, SearcherMixin):
    """ Make sure that we can error message when there is a problem """

    def setUp(self):
        super(BadSearchTest, self).setUp()
        MockSearchEngine.destroy()

    def tearDown(self):
        MockSearchEngine.destroy()
        super(BadSearchTest, self).tearDown()

    def test_search_from_url(self):
        """ ensure that we get the error back when the backend fails """
        searcher = SearchEngine.get_search_engine(TEST_INDEX_NAME)
        searcher.index(
            "courseware_content",
            [
                {
                    "id": "FAKE_ID_1",
                    "content": {
                        "text": "Little Darling, it's been a long long lonely winter"
                    }
                }
            ]
        )
        searcher.index(
            "courseware_content",
            [
                {
                    "id": "FAKE_ID_2",
                    "content": {
                        "text": "Little Darling, it's been a year since sun been gone"
                    }
                }
            ]
        )
        searcher.index("test_doc", [{"id": "FAKE_ID_3", "content": {"text": "Here comes the sun"}}])

        code, results = post_request({"search_string": "sun"})
        self.assertGreater(code, 499)
        self.assertEqual(results["error"], 'An error occurred when searching for "sun"')

        with self.assertRaises(Exception):
            searcher.search(query_string="test search")


@override_settings(SEARCH_ENGINE="search.tests.utils.ErroringIndexEngine")
class BadIndexTest(TestCase, SearcherMixin):
    """ Make sure that we can error message when there is a problem """

    def setUp(self):
        super(BadIndexTest, self).setUp()
        MockSearchEngine.destroy()

    def tearDown(self):
        MockSearchEngine.destroy()
        super(BadIndexTest, self).tearDown()

    def test_search_from_url(self):
        """ ensure that we get the error back when the backend fails """
        searcher = SearchEngine.get_search_engine(TEST_INDEX_NAME)
        with self.assertRaises(Exception):
            searcher.index("courseware_content", [{"id": "FAKE_ID_3", "content": {"text": "Here comes the sun"}}])


@override_settings(SEARCH_ENGINE="search.tests.utils.ForceRefreshElasticSearchEngine")
@override_settings(ELASTIC_FIELD_MAPPINGS={"start_date": {"type": "date"}})
@override_settings(COURSEWARE_INDEX_NAME=TEST_INDEX_NAME)
@ddt.ddt
class ElasticSearchUrlTest(TestCase, SearcherMixin):
    """Elastic-specific tests"""
    def setUp(self):
        super(ElasticSearchUrlTest, self).setUp()
        self.searcher.index(
            "courseware_content",
            [
                {
                    "course": "ABC/DEF/GHI",
                    "id": "FAKE_ID_1",
                    "content": {
                        "text": "It seems like k-means clustering would work in this context."
                    },
                    "test_date": datetime(2015, 1, 1),
                    "test_string": "ABC, It's easy as 123"
                }
            ]
        )
        self.searcher.index(
            "courseware_content",
            [
                {
                    "course": "ABC/DEF/GHI",
                    "id": "FAKE_ID_2",
                    "content": {
                        "text": "It looks like k-means clustering could work in this context."
                    }
                }
            ]
        )
        self.searcher.index(
            "courseware_content",
            [
                {
                    "course": "ABC/DEF/GHI",
                    "id": "FAKE_ID_3",
                    "content": {
                        "text": "It looks like k means something different in this context."
                    }
                }
            ]
        )

    @ddt.data(
        # Quoted phrases
        ('"in this context"', None, 3),
        ('"in this context"', "ABC/DEF/GHI", 3),
        ('"looks like"', None, 2),
        ('"looks like"', "ABC/DEF/GHI", 2),
        # Hyphenated phrases
        ('k-means', None, 3),
        ('k-means', "ABC/DEF/GHI", 3),
    )
    @ddt.unpack
    def test_valid_search(self, query, course_id, result_count):
        code, results = post_request({"search_string": query}, course_id)
        self.assertTrue(199 < code < 300)
        self.assertEqual(results["total"], result_count)

    def test_malformed_query_handling(self):
        # root
        code, results = post_request({"search_string": "\"missing quote"})
        self.assertGreater(code, 499)
        self.assertEqual(results["error"], 'Your query seems malformed. Check for unmatched quotes.')

        # course ID
        code, results = post_request({"search_string": "\"missing quote"}, "ABC/DEF/GHI")
        self.assertGreater(code, 499)
        self.assertEqual(results["error"], 'Your query seems malformed. Check for unmatched quotes.')

        # course discovery
        code, results = post_discovery_request({"search_string": "\"missing quote"})
        self.assertGreater(code, 499)
        self.assertEqual(results["error"], 'Your query seems malformed. Check for unmatched quotes.')



""" Elastic Search implementation for courseware search index """
from __future__ import absolute_import
import copy
import logging

from django.conf import settings
from django.core.cache import cache
from elasticsearch import Elasticsearch, exceptions
from elasticsearch.helpers import bulk, BulkIndexError

import six
from search.api import QueryParseError
from search.search_engine_base import SearchEngine
from search.utils import ValueRange, _is_iterable

# log appears to be standard name used for logger
log = logging.getLogger(__name__)  # pylint: disable=invalid-name

# These are characters that may have special meaning within Elasticsearch.
# We _may_ want to use these for their special uses for certain queries,
# but for analysed fields these kinds of characters are removed anyway, so
# we can safely remove them from analysed matches
RESERVED_CHARACTERS = "+=><!(){}[]^~*:\\/&|?"


def _translate_hits(es_response):
    """ Provide resultset in our desired format from elasticsearch results """

    def translate_result(result):
        """ Any conversion from ES result syntax into our search engine syntax """
        translated_result = copy.copy(result)
        data = translated_result.pop("_source")

        translated_result.update({
            "data": data,
            "score": translated_result["_score"]
        })

        return translated_result

    def translate_facet(result):
        """ Any conversion from ES facet syntax into our search engine sytax """
        terms = {term["term"]: term["count"] for term in result["terms"]}
        return {
            "terms": terms,
            "total": result["total"],
            "other": result["other"],
        }

    results = [translate_result(hit) for hit in es_response["hits"]["hits"]]
    response = {
        "took": es_response["took"],
        "total": es_response["hits"]["total"],
        "max_score": es_response["hits"]["max_score"],
        "results": results,
    }

    if "facets" in es_response:
        response["facets"] = {facet: translate_facet(es_response["facets"][facet]) for facet in es_response["facets"]}

    return response


def _get_filter_field(field_name, field_value):
    """ Return field to apply into filter, if an array then use a range, otherwise look for a term match """
    filter_field = None
    if isinstance(field_value, ValueRange):
        range_values = {}
        if field_value.lower:
            range_values.update({"gte": field_value.lower_string})
        if field_value.upper:
            range_values.update({"lte": field_value.upper_string})
        filter_field = {
            "range": {
                field_name: range_values
            }
        }
    elif _is_iterable(field_value):
        filter_field = {
            "terms": {
                field_name: field_value
            }
        }
    else:
        filter_field = {
            "term": {
                field_name: field_value
            }
        }
    return filter_field


def _process_field_queries(field_dictionary):
    """
    We have a field_dictionary - we want to match the values for an elasticsearch "match" query
    This is only potentially useful when trying to tune certain search operations
    """
    def field_item(field):
        """ format field match as "match" item for elasticsearch query """
        return {
            "match": {
                field: field_dictionary[field]
            }
        }

    return [field_item(field) for field in field_dictionary]


def _process_field_filters(field_dictionary):
    """
    We have a field_dictionary - we match the values using a "term" filter in elasticsearch
    """
    return [_get_filter_field(field, field_value) for field, field_value in field_dictionary.items()]


def _process_filters(filter_dictionary):
    """
    We have a filter_dictionary - this means that if the field is included
    and matches, then we can include, OR if the field is undefined, then we
    assume it is safe to include
    """
    def filter_item(field):
        """ format elasticsearch filter to pass if value matches OR field is not included """
        if filter_dictionary[field] is not None:
            return {
                "or": [
                    _get_filter_field(field, filter_dictionary[field]),
                    {
                        "missing": {
                            "field": field
                        }
                    }
                ]
            }

        return {
            "missing": {
                "field": field
            }
        }

    return [filter_item(field) for field in filter_dictionary]


def _process_exclude_dictionary(exclude_dictionary):
    """
    Based on values in the exclude_dictionary generate a list of term queries that
    will filter out unwanted results.
    """
    # not_properties will hold the generated term queries.
    not_properties = []
    for exclude_property in exclude_dictionary:
        exclude_values = exclude_dictionary[exclude_property]
        if not isinstance(exclude_values, list):
            exclude_values = [exclude_values]
        not_properties.extend([{"term": {exclude_property: exclude_value}} for exclude_value in exclude_values])

    # Returning a query segment with an empty list freaks out ElasticSearch,
    #   so just return an empty segment.
    if not not_properties:
        return {}

    return {
        "not": {
            "filter": {
                "or": not_properties
            }
        }
    }


def _process_facet_terms(facet_terms):
    """ We have a list of terms with which we return facets """
    elastic_facets = {}
    for facet in facet_terms:
        facet_term = {"field": facet}
        if facet_terms[facet]:
            for facet_option in facet_terms[facet]:
                facet_term[facet_option] = facet_terms[facet][facet_option]

        elastic_facets[facet] = {
            "terms": facet_term
        }

    return elastic_facets


class ElasticSearchEngine(SearchEngine):

    """ ElasticSearch implementation of SearchEngine abstraction """

    @staticmethod
    def get_cache_item_name(index_name, doc_type):
        """ name-formatter for cache_item_name """
        return "elastic_search_mappings_{}_{}".format(
            index_name,
            doc_type
        )

    @classmethod
    def get_mappings(cls, index_name, doc_type):
        """ fetch mapped-items structure from cache """
        return cache.get(cls.get_cache_item_name(index_name, doc_type), {})

    @classmethod
    def set_mappings(cls, index_name, doc_type, mappings):
        """ set new mapped-items structure into cache """
        cache.set(cls.get_cache_item_name(index_name, doc_type), mappings)

    @classmethod
    def log_indexing_error(cls, indexing_errors):
        """ Logs indexing errors and raises a general ElasticSearch Exception"""
        indexing_errors_log = []
        for indexing_error in indexing_errors:
            indexing_errors_log.append(str(indexing_error))
        raise exceptions.ElasticsearchException(', '.join(indexing_errors_log))

    def _get_mappings(self, doc_type):
        """
        Interfaces with the elasticsearch mappings for the index
        prevents multiple loading of the same mappings from ES when called more than once

        Mappings format in elasticsearch is as follows:
        {
           "doc_type": {
              "properties": {
                 "nested_property": {
                    "properties": {
                       "an_analysed_property": {
                          "type": "string"
                       },
                       "another_analysed_property": {
                          "type": "string"
                       }
                    }
                 },
                 "a_not_analysed_property": {
                    "type": "string",
                    "index": "not_analyzed"
                 },
                 "a_date_property": {
                    "type": "date"
                 }
              }
           }
        }

        We cache the properties of each doc_type, if they are not available, we'll load them again from Elasticsearch
        """
        # Try loading the mapping from the cache.
        mapping = ElasticSearchEngine.get_mappings(self.index_name, doc_type)

        # Fall back to Elasticsearch
        if not mapping:
            mapping = self._es.indices.get_mapping(
                index=self.index_name,
                doc_type=doc_type,
            ).get(self.index_name, {}).get('mappings', {}).get(doc_type, {})

            # Cache the mapping, if one was retrieved
            if mapping:
                ElasticSearchEngine.set_mappings(
                    self.index_name,
                    doc_type,
                    mapping
                )

        return mapping

    def _clear_mapping(self, doc_type):
        """ Remove the cached mappings, so that they get loaded from ES next time they are requested """
        ElasticSearchEngine.set_mappings(self.index_name, doc_type, {})

    def __init__(self, index=None):
        super(ElasticSearchEngine, self).__init__(index)
        es_config = getattr(settings, "ELASTIC_SEARCH_CONFIG", [{}])
        self._es = getattr(settings, "ELASTIC_SEARCH_IMPL", Elasticsearch)(es_config)
        if not self._es.indices.exists(index=self.index_name):
            self._es.indices.create(index=self.index_name)

    def _check_mappings(self, doc_type, body):
        """
        We desire to index content so that anything we want to be textually searchable(and therefore needing to be
        analysed), but the other fields are designed to be filters, and only require an exact match. So, we want to
        set up the mappings for these fields as "not_analyzed" - this will allow our filters to work faster because
        they only have to work off exact matches
        """

        # Make fields other than content be indexed as unanalyzed terms - content
        # contains fields that are to be analyzed
        exclude_fields = ["content"]
        field_properties = getattr(settings, "ELASTIC_FIELD_MAPPINGS", {})

        def field_property(field_name, field_value):
            """
            Prepares field as property syntax for providing correct mapping desired for field

            Mappings format in elasticsearch is as follows:
            {
               "doc_type": {
                  "properties": {
                     "nested_property": {
                        "properties": {
                           "an_analysed_property": {
                              "type": "string"
                           },
                           "another_analysed_property": {
                              "type": "string"
                           }
                        }
                     },
                     "a_not_analysed_property": {
                        "type": "string",
                        "index": "not_analyzed"
                     },
                     "a_date_property": {
                        "type": "date"
                     }
                  }
               }
            }

            We can only add new ones, but the format is the same
            """
            prop_val = None
            if field_name in field_properties:
                prop_val = field_properties[field_name]
            elif isinstance(field_value, dict):
                props = {fn: field_property(fn, field_value[fn]) for fn in field_value}
                prop_val = {"properties": props}
            else:
                prop_val = {
                    "type": "string",
                    "index": "not_analyzed",
                }

            return prop_val

        new_properties = {
            field: field_property(field, value)
            for field, value in body.items()
            if (field not in exclude_fields) and (field not in self._get_mappings(doc_type).get('properties', {}))
        }

        if new_properties:
            self._es.indices.put_mapping(
                index=self.index_name,
                doc_type=doc_type,
                body={
                    doc_type: {
                        "properties": new_properties,
                    }
                }
            )
            self._clear_mapping(doc_type)

    def index(self, doc_type, sources, **kwargs):
        """
        Implements call to add documents to the ES index
        Note the call to _check_mappings which will setup fields with the desired mappings
        """

        try:
            actions = []
            for source in sources:
                self._check_mappings(doc_type, source)
                id_ = source['id'] if 'id' in source else None
                log.debug("indexing %s object with id %s", doc_type, id_)  # lint-amnesty, pylint: disable=unicode-format-string
                action = {
                    "_index": self.index_name,
                    "_type": doc_type,
                    "_id": id_,
                    "_source": source
                }
                actions.append(action)
            # bulk() returns a tuple with summary information
            # number of successfully executed actions and number of errors if stats_only is set to True.
            _, indexing_errors = bulk(
                self._es,
                actions,
                **kwargs
            )
            if indexing_errors:
                ElasticSearchEngine.log_indexing_error(indexing_errors)
        # Broad exception handler to protect around bulk call
        except Exception as ex:
            # log information and re-raise
            log.exception("error while indexing - %s", str(ex))  # lint-amnesty, pylint: disable=unicode-format-string
            raise

    def remove(self, doc_type, doc_ids, **kwargs):
        """ Implements call to remove the documents from the index """

        try:
            # ignore is flagged as an unexpected-keyword-arg; ES python client documents that it can be used
            # pylint: disable=unexpected-keyword-arg
            actions = []
            for doc_id in doc_ids:
                log.debug("Removing document of type %s and index %s", doc_type, doc_id)  # lint-amnesty, pylint: disable=unicode-format-string
                action = {
                    '_op_type': 'delete',
                    "_index": self.index_name,
                    "_type": doc_type,
                    "_id": doc_id
                }
                actions.append(action)
            bulk(self._es, actions, **kwargs)
        except BulkIndexError as ex:
            valid_errors = [error for error in ex.errors if error['delete']['status'] != 404]

            if valid_errors:
                log.exception("An error occurred while removing documents from the index.")
                raise

    # A few disabled pylint violations here:
    # This procedure takes each of the possible input parameters and builds the query with each argument
    # I tried doing this in separate steps, but IMO it makes it more difficult to follow instead of less
    # So, reasoning:
    #
    #   too-many-arguments: We have all these different parameters to which we
    #       wish to pay attention, it makes more sense to have them listed here
    #       instead of burying them within kwargs
    #
    #   too-many-locals: I think this counts all the arguments as well, but
    #       there are some local variables used herein that are there for transient
    #       purposes and actually promote the ease of understanding
    #
    #   too-many-branches: There's a lot of logic on the 'if I have this
    #       optional argument then...'. Reasoning goes back to its easier to read
    #       the (somewhat linear) flow rather than to jump up to other locations in code
    def search(self,
               query_string=None,
               field_dictionary=None,
               filter_dictionary=None,
               exclude_dictionary=None,
               facet_terms=None,
               exclude_ids=None,
               use_field_match=False,
               **kwargs):  # pylint: disable=too-many-arguments, too-many-locals, too-many-branches, arguments-differ, unicode-format-string
        """  # lint-amnesty, pylint: disable=unicode-format-string
        Implements call to search the index for the desired content.

        Args:
            query_string (str): the string of values upon which to search within the
            content of the objects within the index

            field_dictionary (dict): dictionary of values which _must_ exist and
            _must_ match in order for the documents to be included in the results

            filter_dictionary (dict): dictionary of values which _must_ match if the
            field exists in order for the documents to be included in the results;
            documents for which the field does not exist may be included in the
            results if they are not otherwise filtered out

            exclude_dictionary(dict): dictionary of values all of which which must
            not match in order for the documents to be included in the results;
            documents which have any of these fields and for which the value matches
            one of the specified values shall be filtered out of the result set

            facet_terms (dict): dictionary of terms to include within search
            facets list - key is the term desired to facet upon, and the value is a
            dictionary of extended information to include. Supported right now is a
            size specification for a cap upon how many facet results to return (can
            be an empty dictionary to use default size for underlying engine):

            e.g.
            {
                "org": {"size": 10},  # only show top 10 organizations
                "modes": {}
            }

            use_field_match (bool): flag to indicate whether to use elastic
            filtering or elastic matching for field matches - this is nothing but a
            potential performance tune for certain queries

            (deprecated) exclude_ids (list): list of id values to exclude from the results -
            useful for finding maches that aren't "one of these"

        Returns:
            dict object with results in the desired format
            {
                "took": 3,
                "total": 4,
                "max_score": 2.0123,
                "results": [
                    {
                        "score": 2.0123,
                        "data": {
                            ...
                        }
                    },
                    {
                        "score": 0.0983,
                        "data": {
                            ...
                        }
                    }
                ],
                "facets": {
                    "org": {
                        "total": total_count,
                        "other": 1,
                        "terms": {
                            "MITx": 25,
                            "HarvardX": 18
                        }
                    },
                    "modes": {
                        "total": modes_count,
                        "other": 15,
                        "terms": {
                            "honor": 58,
                            "verified": 44,
                        }
                    }
                }
            }

        Raises:
            ElasticsearchException when there is a problem with the response from elasticsearch

        Example usage:
            .search(
                "find the words within this string",
                {
                    "must_have_field": "mast_have_value for must_have_field"
                },
                {

                }
            )
        """

        log.debug("searching index with %s", query_string)  # lint-amnesty, pylint: disable=unicode-format-string

        elastic_queries = []
        elastic_filters = []

        # We have a query string, search all fields for matching text within the "content" node
        if query_string:
            if six.PY2:
                query_string = query_string.encode('utf-8').translate(None, RESERVED_CHARACTERS)
            else:
                query_string = query_string.translate(query_string.maketrans('', '', RESERVED_CHARACTERS))
            elastic_queries.append({
                "query_string": {
                    "fields": ["content.*"],
                    "query": query_string
                }
            })

        if field_dictionary:
            if use_field_match:
                elastic_queries.extend(_process_field_queries(field_dictionary))
            else:
                elastic_filters.extend(_process_field_filters(field_dictionary))

        if filter_dictionary:
            elastic_filters.extend(_process_filters(filter_dictionary))

        # Support deprecated argument of exclude_ids
        if exclude_ids:
            if not exclude_dictionary:
                exclude_dictionary = {}
            if "_id" not in exclude_dictionary:
                exclude_dictionary["_id"] = []
            exclude_dictionary["_id"].extend(exclude_ids)

        if exclude_dictionary:
            elastic_filters.append(_process_exclude_dictionary(exclude_dictionary))

        query_segment = {
            "match_all": {}
        }
        if elastic_queries:
            query_segment = {
                "bool": {
                    "must": elastic_queries
                }
            }

        query = query_segment
        if elastic_filters:
            filter_segment = {
                "bool": {
                    "must": elastic_filters
                }
            }
            query = {
                "filtered": {
                    "query": query_segment,
                    "filter": filter_segment,
                }
            }

        body = {"query": query}
        if facet_terms:
            facet_query = _process_facet_terms(facet_terms)
            if facet_query:
                body["facets"] = facet_query

        try:
            es_response = self._es.search(
                index=self.index_name,
                body=body,
                **kwargs
            )
        except exceptions.ElasticsearchException as ex:
            message = six.text_type(ex)
            if 'QueryParsingException' in message:
                log.exception("Malformed search query: %s", message)  # lint-amnesty, pylint: disable=unicode-format-string
                raise QueryParseError('Malformed search query.')
            else:
                # log information and re-raise
                log.exception("error while searching index - %s", str(message))  # lint-amnesty, pylint: disable=unicode-format-string
                raise

        return _translate_hits(es_response)



#!/usr/bin/env python
""" Setup to allow pip installs of edx-search module """

from __future__ import absolute_import

from setuptools import setup


def load_requirements(*requirements_paths):
    """
    Load all requirements from the specified requirements files.
    Returns:
        list: Requirements file relative path strings
    """
    requirements = set()
    for path in requirements_paths:
        requirements.update(
            line.split('#')[0].strip() for line in open(path).readlines()
            if is_requirement(line.strip())
        )
    return list(requirements)


def is_requirement(line):
    """
    Return True if the requirement line is a package requirement.
    Returns:
        bool: True if the line is not blank, a comment, a URL, or an included file
    """
    return line and not line.startswith(('-r', '#', '-e', 'git+', '-c'))


setup(
    name='edx-search',
    version='1.3.0',
    description='Search and index routines for index access',
    author='edX',
    author_email='oscm@edx.org',
    url='https://github.com/edx/edx-search',
    license='AGPL',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
    ],
    packages=['search', 'search.tests'],
    install_requires=load_requirements('requirements/base.in')
)
